#include<CoreFoundation/CoreFoundation.h>
#include<CFNetwork/CFSocketStream.h>
#include<map>
#include<pthread.h>
#include"ios_sockets.h"

static ios_watch_callback_t ios_watch_callback=0;
void ios_set_watch_callback(ios_watch_callback_t cb)
{
    ios_watch_callback=cb;
}
void ios_run_watch_callback(int conns, int bytes, int sslbytes)
{
    if (ios_watch_callback)
	ios_watch_callback(conns,bytes,sslbytes);
}
typedef std::map<int,CFReadStreamRef> RefMapType;
static RefMapType RefMap;
static pthread_mutex_t oplock = PTHREAD_MUTEX_INITIALIZER;
int ios_bg_attach_socket(int sock)
{
    if (sock == -1)
	return -1;
    CFReadStreamRef hReadStream = 0;
    CFStreamCreatePairWithSocket (kCFAllocatorDefault, sock,
					    &hReadStream, NULL);
    if (CFReadStreamSetProperty(hReadStream,
					    kCFStreamNetworkServiceType,
					    kCFStreamNetworkServiceTypeVoIP) != TRUE)
    {
	    // An error occured, delete the stream
	    if(hReadStream != NULL)
	    {
		    CFReadStreamClose(hReadStream);
		    CFRelease(hReadStream);
		    hReadStream = NULL;
	    }
	   
	    return -1;
    }
   
    if (CFReadStreamOpen(hReadStream) != TRUE)
    {
	    // An error occured, delete the stream
	    if(hReadStream != NULL)
	    {
		    CFReadStreamClose(hReadStream);
		    CFRelease(hReadStream);
		    hReadStream = NULL;
	    }

	    return -1;
    }
    pthread_mutex_lock(&oplock);
    RefMapType::iterator it = RefMap.find(sock);
    if (it != RefMap.end())
    {
	CFReadStreamClose(it->second);
	CFRelease(it->second);
    }
    RefMap[sock] = hReadStream;
    pthread_mutex_unlock(&oplock);
}

int ios_bg_detach_socket(int sock)
{
    int rv = -1;
    if (sock == -1)
	return -1;
    pthread_mutex_lock(&oplock);
    RefMapType::iterator it = RefMap.find(sock);
    if (it != RefMap.end())
    {
	CFReadStreamClose(it->second);
	CFRelease(it->second);
	RefMap.erase(it);
	rv = 0;
    }
    pthread_mutex_unlock(&oplock);
    return rv;
}
